import logging
from datetime import timedelta
from typing import List

from .album import FullAlbum, SimpleAlbum
from .artist import FullArtist, SimpleArtist
from .audioanalysis import AudioAnalysis
from .audiofeatures import AudioFeatures
from .device import Device
from .http import HTTP
from .oauth.flows import Authenticator, RefreshableFlowMixin
from .object import SpotifyObject
from .pager import CursorBasedPaging, Pager, SearchPager
from .playing import CurrentlyPlaying, CurrentlyPlayingContext
from .playlist import FullPlaylist, SimplePlaylist
from .track import FullTrack, PlaylistTrack, SimpleTrack
from .user import PrivateUser, PublicUser
from .utils import subslice

log = logging.getLogger(__name__)


def get_id(obj):
	if isinstance(obj, SpotifyObject):
		return obj.id
	else:
		return obj


def clamp(limit, minimum):
	return None if limit is None else min(max(limit, 1), minimum)


class Client:
	'''
	Client interface for the API.

	This is the class you should be interfacing with when fetching Spotify objects.

	auth: :class:`Authenticator`
		Authenticator instance used for authenticating with the API.
	'''

	auth: Authenticator
	http: HTTP

	def __init__(self, auth):
		'''
		Creates a Spotify Client instance.

		:param auth: Instance of :class:`Authenticator`
		'''

		self.auth = auth(self)
		self.http = HTTP(self)

	async def __aenter__(self):
		await self.auth.authorize()
		return self

	async def __aexit__(self, exc_type, exc_val, exc_tb):
		await self.close()

	async def authorize(self):
		'''Tell the authenticator to authorize this client.'''

		await self.auth.authorize()

	async def refresh(self):
		'''Tell the authenticator to refresh this client, if applicable.'''

		if not isinstance(self.auth, RefreshableFlowMixin):
			raise TypeError('Authenticator does not support refreshing.')

		await self.auth.refresh()

	async def close(self):
		'''Close this client session.'''

		await self.auth.close()
		await self.http.close()

	def _ensure_market(self, market):
		return self.auth.market if market is None else market

	async def get_player(self, **kwargs) -> CurrentlyPlayingContext:
		'''
		Get a context for what user is currently playing.

		:param kwargs: query params for this request
		:return: :class:`PlayingContext`
		'''

		data = await self.http.get_player(**kwargs)

		return CurrentlyPlayingContext(self, data)

	async def player_currently_playing(self, **kwargs) -> CurrentlyPlaying:
		data = await self.http.player_currently_playing(**kwargs)

		return CurrentlyPlaying(self, data)

	async def get_devices(self) -> List[Device]:
		'''
		Get a list of user devices.

		:return: A list of maximum 20 devices.
		'''

		data = await self.http.get_devices()

		devices = []

		for device_obj in data['devices']:
			devices.append(Device(self, device_obj))

		return devices

	async def player_next(self, device=None):
		'''
		Skip to the next track in a player.

		:param device: :class:`Device` or Spotify ID.
		'''

		await self.http.player_next(device_id=get_id(device))

	async def player_prev(self, device=None):
		'''
		Play the previous track.

		:param device: :class:`Device` or Spotify ID.
		'''

		await self.http.player_prev(device_id=get_id(device))

	async def player_play(self, device=None, **kwargs):
		'''
		Start playback on device.

		:param device: :class:`Device` or Spotify ID.
		:param kwargs: body params of the request.

		Example:

		.. code-block:: py

			player_play(
				context_uri='spotify:album:6xKK037rfCf2f6gf30SpvL',
				offset=dict(uri='spotify:track:2beor6qrB0XJxW1CM6X9x2'),
				position_ms=98500
			)
		'''

		await self.http.player_play(device_id=get_id(device), **kwargs)

	async def player_pause(self, device=None):
		'''
		Stop playback on a device.

		:param device: :class:`Device` or Spotify ID.
		'''

		await self.http.player_pause(device_id=get_id(device))

	async def player_seek(self, time, device=None):
		'''
		Seek to a point in the currently playing track.

		:param time: timedelta object or milliseconds (integer)
		:param device: :class:`Device` or Spotify ID.
		'''

		if isinstance(time, timedelta):
			time = time.days * (1000 * 60 * 60 * 24) + time.seconds * 1000 + time.microseconds // 1000
		await self.http.player_seek(time, get_id(device))

	async def player_repeat(self, state, device=None):
		'''
		Set player repeat mode.

		:param str state: Can be 'track', 'context' or 'off'.
		:param device: :class:`Device` or Spotify ID.
		'''

		await self.http.player_repeat(state, get_id(device))

	async def player_volume(self, volume, device=None):
		'''
		Set player volume.

		:param int volume: Value from 0 to 100.
		:param device: :class:`Device` or Spotify ID.
		'''

		await self.http.player_volume(volume, get_id(device))

	async def player_shuffle(self, state, device=None):
		'''
		Set player shuffle mode.

		:param bool state: Shuffle mode state.
		:param device: :class:`Device` or Spotify ID.
		'''

		await self.http.player_shuffle(state, get_id(device))

	async def search(self, *types, q, limit=20, market=None, offset=None, include_external=None) -> dict:
		'''
		Searches for tracks, artists, albums and/or playlists.

		:param types: One or more of the strings ``track``, ``album``, ``artist``, ``playlist`` or the class equivalents.
		:param str q: The search query. See Spotifys' query construction guide `here. <https://developer.spotify.com/documentation/web-api/reference/search/search/>`_
		:param int limit: How many results of each type to return.
		:param market: ISO-3166-1_ country code or the string ``from_token``.
		:param offset: Where to start the pagination.
		:param include_external: If this is equal to ``audio``, the specified the response will include any relevant audio content that is hosted externally.

		:return: A dict with a key for each type, whose values are a list of instances.
		'''

		actual_types = []
		for type in types:
			if isinstance(type, (SimpleTrack, SimpleAlbum, SimpleArtist, SimplePlaylist)):
				actual_types.append(type._type)
			elif isinstance(type, str):
				actual_types.append(type.lower())
			else:
				raise ValueError('Unknown type: %s' % str(type))

		data = await self.http.search(
			q, ','.join(actual_types),
			market=market,
			limit=clamp(limit, 50),
			offset=offset,
			include_external=include_external
		)

		results = {}

		if 'tracks' in data:
			results['tracks'] = []
			async for track_obj in SearchPager(self.http, data, 'tracks', limit):
				results['tracks'].append(SimpleTrack(self, track_obj))

		if 'albums' in data:
			results['albums'] = []
			async for album_obj in SearchPager(self.http, data, 'albums', limit):
				results['albums'].append(SimpleAlbum(self, album_obj))

		if 'artists' in data:
			results['artists'] = []
			async for artist_obj in SearchPager(self.http, data, 'artists', limit):
				results['artists'].append(FullArtist(self, artist_obj))

		if 'playlists' in data:
			results['playlists'] = []
			async for artist_obj in SearchPager(self.http, data, 'playlists', limit):
				results['playlists'].append(SimplePlaylist(self, artist_obj))

		return results

	async def search_tracks(self, q, limit=20, market=None, offset=None, include_external=None) -> List[SimpleTrack]:
		'''
		Alias for ``Client.search('track', ...)``

		:return: List[:class:`SimpleTrack`]
		'''

		results = await self.search(
			'track', q=q, market=market, limit=limit, offset=offset, include_external=include_external
		)
		return results['tracks']

	async def search_artists(self, q, limit=20, market=None, offset=None, include_external=None) -> List[FullArtist]:
		'''
		Alias for ``Client.search('artist, ...)``

		:return: List[:class:`FullArtist`]
		'''

		results = await self.search(
			'artist', q=q, market=market, limit=limit, offset=offset, include_external=include_external
		)
		return results['artists']

	async def search_albums(self, q, limit=20, market=None, offset=None, include_external=None) -> List[SimpleAlbum]:
		'''
		Alias for ``Client.search('album', ...)``

		:return: List[:class:`SimpleAlbum`]
		'''

		results = await self.search(
			'album', q=q, market=market, limit=limit, offset=offset, include_external=include_external
		)
		return results['albums']

	async def search_playlists(self, q, limit=20, market=None, offset=None, include_external=None) -> List[SimplePlaylist]:
		'''
		Alias for ``Client.search('playlist', ...)``

		:return: List[:class:`SimplePlaylist`]
		'''

		results = await self.search(
			'playlist', q=q, market=market, limit=limit, offset=offset, include_external=include_external
		)
		return results['playlists']

	async def search_track(self, q=None) -> SimpleTrack:
		'''
		Returns the top track for the query.

		:return: :class:`SimpleTrack` or None
		'''

		results = await self.search_tracks(q=q, limit=1)
		return results[0] if results else None

	async def search_artist(self, q=None) -> FullArtist:
		'''
		Returns the top artist for the query.

		:return: :class:`SimpleArtist` or None
		'''

		results = await self.search_artists(q=q, limit=1)
		return results[0] if results else None

	async def search_album(self, q=None) -> SimpleAlbum:
		'''
		Returns the top album for the query.

		:return: :class:`SimpleAlbum`
		'''

		results = await self.search_albums(q=q, limit=1)
		return results[0] if results else None

	async def search_playlist(self, q=None) -> SimplePlaylist:
		'''
		Returns the top playlist for the query.

		:return: :class:`SimplePlaylist`
		'''

		results = await self.search_playlists(q=q, limit=1)
		return results[0] if results else None

	async def get_me(self) -> PrivateUser:
		'''
		Gets the current user.

		:return: A :class:`PrivateUser` instance of the current user.
		'''

		data = await self.http.get_me()
		return PrivateUser(self, data)

	async def get_me_top_tracks(self, limit=20, offset=None, time_range=None) -> List[SimpleTrack]:
		'''
		Gets the top tracks of the current user.

		Requires scope ``user-top-read``.

		:param int limit: How many tracks to return. Maximum is 50.
		:param int offset: The index of the first result to return.
		:param str time_range: The time period for which data are selected to form a top.

		Valid values for ``time_range``
		  - ``long_term`` (calculated from several years of data and including all new data as it becomes available),
		  - ``medium_term`` (approximately last 6 months),
		  - ``short_term`` (approximately last 4 weeks).

		:return: List[:class:`SimpleTrack`]
		'''

		data = await self.http.get_me_top_tracks(limit=clamp(limit, 50), offset=offset, time_range=time_range)

		tracks = []
		for track_obj in data['items']:
			tracks.append(SimpleTrack(self, track_obj))

		return tracks

	async def get_me_top_artists(self, limit=20, offset=0, time_range='medium_term') -> List[SimpleArtist]:
		'''
		Get the top artists of the current user.

		:param int limit: How many artists to return. Maximum is 50.
		:param int offset: The index of the first result to return.
		:param str time_range: The time period for which data are selected to form a top.

		Valid values for ``time_range``
		  - ``long_term`` (calculated from several years of data and including all new data as it becomes available),
		  - ``medium_term`` (approximately last 6 months),
		  - ``short_term`` (approximately last 4 weeks).

		:return: List[:class:`SimpleArtist`]
		'''

		data = await self.http.get_me_top_artists(limit=clamp(limit, 50), offset=offset, time_range=time_range)

		artists = []
		for artist_obj in data['items']:
			artists.append(SimpleArtist(self, artist_obj))

		return artists

	async def get_user(self, user_id) -> PublicUser:
		'''
		Get a user.

		:param str user_id: Spotify ID of user.
		:return: A :class:`PublicUser` instance.
		'''

		data = await self.http.get_user(user_id)

		return PublicUser(self, data)

	async def create_playlist(self, user, name, public=False, collaborative=False, description=None) -> FullPlaylist:
		'''
		Create a new playlist.

		:param user: :class:`User` instance or Spotify ID.
		:param str name: Name of the new playlist.
		:param str description: Description of the new playlist.
		:param bool public: Whether the playlist should be public.
		:param bool collaborative: Whether the playlist should be collaborative (anyone can edit it).
		:return: A :class:`FullPlaylist` instance.
		'''

		data = await self.http.create_playlist(
			get_id(user), name,
			public=public,
			collaborative=collaborative,
			description=description
		)

		playlist = FullPlaylist(self, data)
		playlist._tracks = dict()

		return playlist

	async def edit_playlist(self, playlist, name=None, description=None, public=None, collaborative=None):
		'''
		Edit a playlist.

		:param playlist: :class:`Playlist` instance or Spotify ID.
		:param str name: New name of the playlist.
		:param str description: New description of the playlist.
		:param bool public: New public state of the playlist.
		:param bool collaborative: New collaborative state of the playlist.
		'''

		await self.http.edit_playlist(
			playlist_id=get_id(playlist),
			name=name,
			description=description,
			public=public,
			collaborative=collaborative
		)

	async def playlist_add_tracks(self, playlist, *tracks, position=None):
		'''
		Add several tracks to a playlist.

		:param playlist: :class:`Playlist` instance or Spotify ID.
		:param tracks: List of Spotify IDs or :class:`Track` instance (or a mix).
		:param int position: Position in the playlist to insert tracks.
		'''

		playlist = get_id(playlist)

		uris = []

		for track in tracks:
			track = get_id(track)

			if not track.startswith('spotify:track:'):
				track = 'spotify:track:' + track

			uris.append(track)

		for chunk in subslice(uris, 100):
			await self.http.playlist_add_tracks(playlist, uris=chunk, position=position)

	async def get_playlist(self, playlist_id) -> FullPlaylist:
		'''
		Get a pre-existing playlist.

		:param str playlist_id: Spotify ID of the playlist.
		:return: :class:`FullPlaylist` instance.
		'''

		data = await self.http.get_playlist(playlist_id)

		playlist = FullPlaylist(self, data)

		return playlist

	async def get_playlist_tracks(self, playlist) -> List[PlaylistTrack]:
		'''
		Get tracks from a playlist.

		:param playlist: :class:`Playlist` instance or Spotify ID.
		:return: List[:class:`PlaylistTrack`]
		'''

		playlist = get_id(playlist)

		data = await self.http.get_playlist_tracks(playlist)

		tracks = []

		async for track in Pager(self.http, data):
			tracks.append(PlaylistTrack(self, track))

		return tracks

	async def get_user_playlists(self, user) -> List[SimplePlaylist]:
		'''
		Get a list of attainable playlists a user owns.

		:param user: :class:`User` instance or Spotify ID.
		:return: List[:class:`SimplePlaylist`]
		'''

		user = get_id(user)

		data = await self.http.get_user_playlists(user)

		playlists = []

		async for playlist_obj in Pager(self.http, data):
			if playlist_obj is None:
				playlists.append(None)
			else:
				playlists.append(SimplePlaylist(self, playlist_obj))

		return playlists

	async def get_track(self, track_id) -> FullTrack:
		'''
		Get a track.

		:param str track_id: Spotify ID of track.
		:return: :class:`FullTrack` instance.
		'''

		data = await self.http.get_track(track_id)
		return FullTrack(self, data)

	async def get_tracks(self, *track_ids) -> List[FullTrack]:
		'''
		Get several tracks.

		:param str track_ids: List of track Spotify IDs.
		:return: List[:class:`FullTrack`]
		'''

		tracks = []

		for chunk in subslice(track_ids, 50):
			data = await self.http.get_tracks(','.join(get_id(obj) for obj in chunk))

			for track_obj in data['tracks']:
				if track_obj is None:
					tracks.append(None)
				else:
					tracks.append(FullTrack(self, track_obj))

		return tracks

	async def get_audio_features(self, track) -> AudioFeatures:
		'''
		Get Audio Features of a track.

		:param track: :class:`Track` instance or Spotify ID.
		:return: :class:`AudioFeatures`
		'''

		track = get_id(track)

		data = await self.http.get_audio_features(track)

		return AudioFeatures(self, data)

	async def get_audio_analysis(self, track) -> AudioAnalysis:
		'''
		Get Audio Analysis of a track.

		:param track: :class:`Track` instance or Spotify ID.
		:return: :class:`AudioAnalysis`
		'''

		track = get_id(track)

		data = await self.http.get_audio_analysis(track)

		return AudioAnalysis(self, data)

	async def get_artist(self, artist_id) -> FullArtist:
		'''
		Get an artist.

		:param str artist_id: Spotify ID of artist.
		:return: :class:`FullArtist` instance.
		'''

		data = await self.http.get_artist(artist_id)

		return FullArtist(self, data)

	async def get_artist_albums(self, artist_id, limit=20, include_groups=None, country=None, offset=None) -> List[SimpleAlbum]:
		'''
		Get an artist's albums.

		.. note::
		   This endpoint does *not* return the track objects for each album. If you need those, you have to
		   fetch them manually afterwards.

		:param str artist_id: Spotify ID of artist.
		:param int limit: How many albums to return.
		:param kwargs: other query params for this method
		:return: List[:class:`SimpleAlbum`]
		'''

		albums = []

		data = await self.http.get_artist_albums(
			artist_id, include_groups=include_groups, country=country, limit=clamp(limit, 50), offset=offset
		)

		async for album_obj in Pager(self.http, data, limit):
			albums.append(SimpleAlbum(self, album_obj))

		return albums

	async def get_artists(self, *artist_ids) -> List[FullArtist]:
		'''
		Get several artists.

		:param artist_ids: List of artist Spotify IDs.
		:return: List[:class:`FullArtist`]
		'''

		artists = []

		for chunk in subslice(artist_ids, 2):
			data = await self.http.get_artists(','.join(get_id(obj) for obj in chunk))

			for artist_obj in data['artists']:
				if artist_obj is None:
					artists.append(None)
				else:
					artists.append(FullArtist(self, artist_obj))

		return artists

	async def get_artist_top_tracks(self, artist, market=None) -> List[FullTrack]:
		'''
		Returns the top tracks for an artist.

		:param artist: :class:`Artist` instance or Spotify ID.
		:param market: ISO-3166-1_ country code. Leave blank to let the library auto-resolve this.
		:return: A list of maximum 10 :class:`FullTrack` instances.
		'''

		artist_id = get_id(artist)
		market = self._ensure_market(market)

		data = await self.http.get_artist_top_tracks(artist_id, market=market)

		tracks = []

		for track_obj in data['tracks']:
			tracks.append(FullTrack(self, track_obj))

		return tracks

	async def get_artist_related_artists(self, artist) -> List[FullArtist]:
		'''
		Get an artists related artists.

		:param artist: :class:`Artist` or Spotify ID.
		:return: A list of maximum 20 :class:`FullArtist` instances.
		'''

		artist_id = get_id(artist)

		data = await self.http.get_artist_related_artists(artist_id)

		artists = []

		for artist_obj in data['artists']:
			artists.append(FullArtist(self, artist_obj))

		return artists

	async def get_album(self, album_id, market=None) -> FullAlbum:
		'''
		Get an album.

		:param str album_id: Spotify ID of album.
		:param market: ISO-3166-1_ country code.
		:return: :class:`FullAlbum` instance.
		'''

		data = await self.http.get_album(album_id, market=market)

		album = FullAlbum(self, data)

		return album

	async def get_albums(self, *album_ids, market=None) -> List[FullAlbum]:
		'''
		Get several albums.

		:param str album_ids: Spotify ID of album.
		:param market: ISO-3166-1_ country code.
		:return: List[:class:`FullAlbum`]
		'''

		albums = []

		for chunk in subslice(album_ids, 20):
			data = await self.http.get_albums(','.join(get_id(obj) for obj in chunk), market=market)

			for album_obj in data['albums']:
				if album_obj is None:
					albums.append(None)
				else:
					albums.append(FullAlbum(self, album_obj))

		return albums

	async def get_album_tracks(self, album, limit=20, offset=None, market=None) -> List[SimpleTrack]:
		'''
		Get tracks from an album.

		:param album: :class:`Album` or Spotify ID of album.
		:param limit: How many tracks to fetch.
		:param offset: What pagination offset to start from.
		:param market: ISO-3166-1_ country code.
		:return: List[:class:`SimpleTrack`]
		'''

		album = get_id(album)

		# assuming the limit is 50 for now
		data = await self.http.get_album_tracks(album, limit=clamp(limit, 50), offset=offset, market=market)

		tracks = []

		async for track_obj in Pager(self.http, data, limit):
			tracks.append(SimpleTrack(self, track_obj))

		return tracks

	async def get_followed_artists(self, limit=20, after=None) -> List[SimpleArtist]:
		'''
		Get user's followed artists

		:param int limit: The maximum number of items to return. Default - infinity
		:param after: What artist ID to start the fetching from.
		:return: List[:class:`SimpleArtist`]
		'''

		# assuming the limit is 50 for now
		data = await self.http.get_followed_artists(type='artist', limit=clamp(limit, 50), after=after)

		artists = []

		async for artist_obj in CursorBasedPaging(self.http, data, 'artists', limit):
			artists.append(SimpleArtist(self, artist_obj))

		return artists

	async def following(self, type, *ids, limit=None, after=None):
		'''
		Follow artists or users.

		:param str type: The ID type: either artist or user.
		:param str ids: Spotify ID of the artist or the user.
		:param limit: Maximum number of items to return.
		:param after: The last artist ID retrieved from the previous request.
		'''

		# not sure what this endpoint is doing, it doesn't take a limit at least. will look at it some other time
		raise NotImplementedError

		for chunk in subslice(ids, 50):
			await self.http.following(type=type, ids=','.join(get_id(obj) for obj in chunk), limit=clamp(limit, 50), after=after)



from datetime import datetime, timedelta

from .audioanalysis import AudioAnalysis
from .audiofeatures import AudioFeatures
from .mixins import ArtistMixin, ExternalIDMixin, ExternalURLMixin
from .object import SpotifyObject


class _BaseTrack(SpotifyObject, ExternalURLMixin, ArtistMixin):
	_type = 'track'

	def __init__(self, client, data):
		super().__init__(client, data)

		ExternalURLMixin.__init__(self, data)
		ArtistMixin.__init__(self, data)

		self.available_markets = data.pop('available_markets', None)
		self.disc_number = data.pop('disc_number')
		self.explicit = data.pop('explicit')
		self.preview_url = data.pop('preview_url')
		self.track_number = data.pop('track_number')
		self.is_local = data.pop('is_local')

		self.duration = timedelta(milliseconds=data.pop('duration_ms'))

	def avaliable_in(self, market):
		'''
		Check if track is available in a market.

		:param market: ISO-3166-1_ value.
		:return:
		'''
		return market in self.available_markets

	async def audio_features(self) -> AudioFeatures:
		'''
		Get 'Audio Features' of the track.
		
		:param track: :class:`Track` instance or Spotify ID of track.
		:return: :class:`AudioFeatures`
		'''

		return await self._client.get_audio_features(self.id)

	async def audio_analysis(self) -> AudioAnalysis:
		'''
		Get 'Audio Analysis' of the track.
		
		:param track: :class:`Track` instance or Spotify ID of track.
		:return: :class:`AudioAnalysis`
		'''
		return await self._client.get_audio_analysis(self.id)


class SimpleTrack(_BaseTrack):
	'''
	Represents a Track object.

	id: str
		Spotify ID of the track.
	name: str
		Name of the track.
	artists: List[:class:`Artist`]
		List of artists that appear on the track.
	images: List[:class:`Image`]
		List of associated images, such as album cover in different sizes.
	uri: str
		Spotify URI of the album.
	link: str
		Spotify URL of the album.
	type: str
		Plaintext string of object type: ``track``.
	available_markets: List[str] or None
		Markets where the album is available in ISO-3166-1_ form.
	disc_number: int
		What disc the track appears on. Usually ``1`` unless there are several discs in the album.
	duration: `timedelta <https://docs.python.org/3/library/datetime.html#timedelta-objects>`_
		timedelta instance representing the length of the track.
	explicit: bool
		Whether the track is explicit or not.
	external_urls: dict
		Dictionary that maps type to url.
	is_playable: bool
		tbc
	linked_from: :class:`LinkedTrack`
		tbc
	restrictions: restrictions object
		tbc
	preview_url: str
		An URL to a 30 second preview (MP3) of the track.
	track_number: int
		The number of the track on the album.
	is_local: bool
		Whether the track is from a local file.
	'''


class FullTrack(_BaseTrack, ExternalIDMixin):
	'''
	Represents a complete Track object.
	
	This type has some additional attributes not existent in :class:`SimpleTrack`.
	
	album: :class:`SimpleAlbum`
		An instance of the album the track appears on.
	popularity: int
		An indicator of the popularity of the track, 0 being least popular and 100 being the most.
	external_ids: dict
		Dictionary of external IDs.
	'''

	def __init__(self, client, data):
		super().__init__(client, data)

		ExternalIDMixin.__init__(self, data)

		self.popularity = data.pop('popularity')

		from .album import SimpleAlbum
		self.album = SimpleAlbum(client, data.pop('album'))


class PlaylistTrack(FullTrack):
	'''
	Represents a Track object originating from a playlist.
	
	This type has some additional attributes not existent in :class:`SimpleTrack` or :class:`FullTrack`.
	
	added_at: `datetime <https://docs.python.org/3/library/datetime.html#module-datetime>`_
		Indicates when the track was added to the playlist.
	added_by: User object
		Indicates who added the track to the playlist. The information provided from the API is not enough to instantiate a PublicUser object, so it's a plain copy of the returned json object.
	'''

	def __init__(self, client, data):
		super().__init__(client, data.pop('track'))

		self.added_at = datetime.strptime(data.pop('added_at'), "%Y-%m-%dT%H:%M:%SZ")
		self.added_by = data.pop('added_by')



import logging

from .http import Route
from .image import Image
from .object import SpotifyObject
from .pager import Pager

log = logging.getLogger(__name__)


class ArtistMixin:
	def __init__(self, data):
		from .artist import SimpleArtist

		artists = data.pop('artists')
		self.artists = []

		for art in artists:
			self.artists.append(SimpleArtist(self._client, art))


class ImageMixin:
	def __init__(self, data):
		images = data.pop('images', None)
		self.images = []

		if images is not None:
			for img in images:
				self.images.append(Image(img))


class ExternalIDMixin:
	def __init__(self, data):
		external_ids = data.pop('external_ids')
		self.external_ids = {}

		for key, value in external_ids.items():
			self.external_ids[key] = value


class ExternalURLMixin:
	def __init__(self, data):
		external_urls = data.pop('external_urls')
		self.external_urls = {}

		self.link = external_urls.get('spotify', None)
		for site, url in external_urls.items():
			self.external_urls[site] = url


def valid_item(item):
	if item is None:
		return False
	# maybe explicitly check for self._track_class?
	if 'track' in item and item['track'] is None:
		return False
	return True


class TrackMixin:
	def __init__(self, data):
		tracks = data.pop('tracks', None)
		if tracks is None:
			return

		items = tracks.pop('items', None)
		if items is None:
			return

		self.tracks = []

		for item in items:
			if valid_item(item):
				self._add_track(item)

		self.__total = tracks.get('total')
		self.__limit = tracks.get('limit')

	async def __aiter__(self):
		'''Create a pager and iterate all tracks in this object. Also updates the ``tracks`` cache.'''

		self.tracks = []

		# get new pager
		r = Route('GET', '{0}s/{1}/tracks'.format(self._type, self.id), offset=0, limit=self.__limit)
		pager_data = await self._client.http.request(r)

		self.__total = pager_data.get('total')

		async for item in Pager(self._client.http, pager_data):
			if valid_item(item):
				yield self._add_track(item)
			else:
				# if it's not a valid item it shouldn't count towards the total track count goal
				self.__total -= 1

	async def fill(self):
		'''Update this objects ``tracks`` cache.'''

		async for track in self:
			pass

	def is_filled(self):
		'''Whether this object contains as many tracks as advertised by the previous pager.'''

		return len(self.tracks) >= self.__total

	def has_track(self, track):
		'''Check if this object has a track.'''

		if isinstance(track, SpotifyObject):
			track = track.id
		return any(track == t.id for t in self.tracks)

	def _add_track(self, track_data):
		track = self._track_class(self._client.http, track_data)
		self.tracks.append(track)
		return track



from datetime import datetime, timedelta


class AuthenticationResponse:
	access_token: str
	token_type: str
	expires_in: int
	created_at: datetime
	expires_at: datetime

	def __init__(self, data):
		self.access_token = data.get('access_token')
		self.token_type = data.get('token_type')
		self.expires_in = data.get('expires_in')

		# for convenience (and RefreshableMixin timing), not really
		# part of a standard or API response
		self.created_at = datetime.utcnow()
		self.expires_at = self.created_at + timedelta(seconds=self.expires_in)

	@classmethod
	def from_data(cls, data):
		self = cls(data)
		self.created_at = datetime.fromisoformat(data.get('created_at'))
		self.expires_at = datetime.fromisoformat(data.get('expires_at'))
		return self

	@property
	def header(self):
		return dict(Authorization='%s %s' % (self.token_type, self.access_token))

	def to_dict(self):
		return dict(
			access_token=self.access_token,
			token_type=self.token_type,
			expires_in=self.expires_in,
			created_at=self.created_at.isoformat(),
			expires_at=self.expires_at.isoformat(),
		)

	def is_expired(self):
		return self.expires_at < datetime.utcnow()

	def seconds_until_expire(self):
		return (self.expires_at - datetime.utcnow()).total_seconds()


class AuthorizationCodeFlowResponse(AuthenticationResponse):
	scope: str
	refresh_token: str

	def __init__(self, data):
		super().__init__(data)
		self.scope = data.get('scope')
		self.refresh_token = data.get('refresh_token', None)  # can be None during refresh!

	@classmethod
	def from_data(cls, data):
		self = cls(data)
		self.scope = data.get('scope')
		self.refresh_token = data.get('refresh_token') # however should never be None from a stored copy
		return self

	def to_dict(self):
		data = super().to_dict()

		data.update(
			scope=self.scope,
			refresh_token=self.refresh_token,
		)

		return data




from .mixins import ExternalURLMixin, ImageMixin
from .object import SpotifyObject


class _BaseUser(SpotifyObject, ExternalURLMixin, ImageMixin):
	_type = 'user'

	def __init__(self, client, data):
		super().__init__(client, data)

		ExternalURLMixin.__init__(self, data)
		ImageMixin.__init__(self, data)

		self.display_name = data.pop('display_name')
		self.name = self.display_name

		followers = data.get('followers', None)
		self.follower_count = None if followers is None else followers['total']

	async def playlists(self):
		'''
		Get the users playlists.
		
		Alias of :meth:`Client.get_user_playlists`
		'''

		return await self._client.get_user_playlists(self.id)


class PublicUser(_BaseUser):
	'''
	Represents a User object.

	id: str
		Spotify ID of the user.
	name: str
		Name of the user. Also aliased to the ``display_name`` attribute.
	images: List[:class:`Image`]
		List of associated images, such as the users profile picture.
	uri: str
		Spotify URI of the user.
	link: str
		Spotify URL of the user.
	follower_count: int or None
		Follower count of the user.
	external_urls: dict
		Dictionary that maps type to url.
	'''


class PrivateUser(_BaseUser):
	'''
	Represents a private User object, usually fetched through the ``me`` endpoint.

	This type has some additional attributes not existent in :class:`PublicUser`.

	country: str
		ISO-3166-1_ code of users country.
	email: str
		Email of user. Please do not this email is note necessarily verified by Spotify.
	product: str
		Users Spotify subscription level, could be ``free``, ``open`` or ``premium``. ``free`` and ``open`` are synonyms.
	'''

	def __init__(self, client, data):
		super().__init__(client, data)

		# this seems to have been removed from the api listing?
		#self.birthdate = data.pop('birthdate', None)

		self.country = data.pop('country', None)
		self.email = data.pop('email', None)
		self.product = data.pop('product', None)

	async def top_tracks(self, limit=20, offset=None, time_range=None):
		'''
		Gets the top tracks of the current user.

		Requires scope ``user-top-read``.

		:param int limit: How many tracks to return. Maximum is 50.
		:param int offset: The index of the first result to return.
		:param str time_range: The time period for which data are selected to form a top.

		Valid values for ``time_range``
		  - ``long_term`` (calculated from several years of data and including all new data as it becomes available),
		  - ``medium_term`` (approximately last 6 months),
		  - ``short_term`` (approximately last 4 weeks).

		:return: List[:class:`SimpleTrack`]
		'''

		return await self._client.get_me_top_tracks(limit=limit, offset=offset, time_range=time_range)

	async def top_artists(self, limit=20, offset=None, time_range=None):
		'''
		Get the top artists of the current user.

		:param int limit: How many artists to return. Maximum is 50.
		:param int offset: The index of the first result to return.
		:param str time_range: The time period for which data are selected to form a top.

		Valid values for ``time_range``
		  - ``long_term`` (calculated from several years of data and including all new data as it becomes available),
		  - ``medium_term`` (approximately last 6 months),
		  - ``short_term`` (approximately last 4 weeks).

		:return: List[:class:`SimpleArtist`]
		'''

		return await self._client.get_me_top_artists(limit=limit, offset=offset, time_range=time_range)

	async def create_playlist(self, name, public=False, collaborative=False, description=None):
		'''
		Create a new playlist.
		
		:param str name: Name of the new playlist.
		:param str description: Description of the new playlist.
		:param bool public: Whether the playlist should be public.
		:param bool collaborative: Whether the playlist should be collaborative (anyone can edit it).
		:return: A :class:`FullPlaylist` instance.
		'''

		return await self._client.create_playlist(
			user=self.id,
			name=name,
			public=public,
			collaborative=collaborative,
			description=description
		)



from .mixins import ExternalURLMixin, ImageMixin
from .object import SpotifyObject


class _BaseArtist(SpotifyObject, ExternalURLMixin):
	_type = 'artist'

	def __init__(self, client, data):
		super().__init__(client, data)

		ExternalURLMixin.__init__(self, data)

	async def top_tracks(self, market=None):
		'''
		Returns this artists top tracks.

		:param market: Market to find tracks for. Auto-resolved by the library if left blank.
		:return: List[:class:`FullTrack`]
		'''

		return await self._client.get_artist_top_tracks(self.id, market=market)

	async def related_artists(self):
		'''
		Get related artists. Maximum of 20 artists.

		:return: List[:class:`FullArtist`]
		'''

		return await self._client.get_artist_related_artists(self.id)

	async def albums(self, limit=20, include_groups=None, country=None, offset=None):
		'''
		Get artists albums

		:return: List[:class:`SimpleAlbum`]
		'''

		return await self._client.get_artist_albums(
			self.id,
			include_groups=include_groups,
			country=country,
			limit=limit,
			offset=offset
		)


class SimpleArtist(_BaseArtist):
	'''
	Represents an Artist object.

	id: str
		Spotify ID of the artist.
	name: str
		Name of the artist.
	uri: str
		Spotify URI of the artist.
	link: str
		Spotify URL of the artist.
	external_urls: dict
		Dictionary that maps type to url.
	'''


class FullArtist(_BaseArtist, ImageMixin):
	'''
	Represents a complete Artist object.
	
	This type has some additional attributes not existent in :class:`SimpleArtist`.
	
	follow_count: int
		Follow count of the artist.
	genres: List[str]
		Genres associated with the artist.
	popularity: int
		An indicator of the popularity of the track, 0 being least popular and 100 being the most.
	images: List[:class:`Image`]
		List of associated images.
	'''

	def __init__(self, client, data):
		super().__init__(client, data)
		
		ImageMixin.__init__(self, data)

		self.follower_count = data['followers']['total']
		self.genres = data.pop('genres')
		self.popularity = data.pop('popularity')



from collections import namedtuple
from datetime import timedelta

from .object import SpotifyObject

TimeInterval = namedtuple('TimeInterval', ('start', 'duration', 'confidence'))

Section = namedtuple(
	'Section', (
		'start', 'duration', 'confidence',
		'loudness', 'tempo', 'tempo_confidence',
		'key', 'key_confidence', 'mode',
		'mode_confidence', 'time_signature', 'time_signature_confidence'
	)
)

Segment = namedtuple(
	'Segment', (
		'start', 'duration', 'confidence',
		'loudness_start', 'loudness_max', 'loudness_max_time',
		'loudness_end', 'pitches', 'timbre'
	)
)


Track = namedtuple(
	'Track', (
		'duration', 'sample_md5', 'offset_seconds',
		'window_seconds', 'analysis_sample_rate', 'analysis_channels',
		'end_of_fade_in', 'start_of_fade_out', 'loudness',
		'tempo', 'tempo_confidence', 'time_signature',
		'time_signature_confidence', 'key', 'key_confidence',
		'mode', 'mode_confidence', 'codestring',
		'code_version', 'echoprintstring', 'echoprint_version',
		'synchstring', 'synch_version', 'rhythmstring',
		'rhythm_version', 'num_samples'
	)
)


class AudioAnalysis(SpotifyObject):
	'''
	Represents an Audio Analysis object.

	This page only skims the details on this object. *Please* read the official Spotify documentation `here <https://developer.spotify.com/documentation/web-api/reference/tracks/get-audio-analysis/>`_.

	bars: List[TimeInterval]
		The time intervals of the bars throughout the track. A bar (or measure) is a segment of time defined as a given number of beats.
	beats: List[TimeInterval]
		The time intervals of beats throughout the track. A beat is the basic time unit of a piece of music; for example, each tick of a metronome.
	sections: List[Section]
		List of sections of the track. Sections are defined by large variations in rhythm or timbre, e.g. chorus, verse, bridge, guitar solo, etc.
	segments: List[Segment]
		List of audio segments of the track. Audio segments attempts to subdivide a song into many segments, with each segment containing a roughly consistent sound throughout its duration.
	tatums: List[TimeInterval]
		The time intervals of tatums throughout the track. A tatum represents the lowest regular pulse train that a listener intuitively infers from the timing of perceived musical events (segments).
	'''

	_type = 'audio_analysis'

	def __init__(self, client, data):
		super().__init__(client, data)

		self.bars = [
			TimeInterval(
				start=timedelta(seconds=bar.pop('start')),
				duration=timedelta(seconds=bar.pop('duration')),
				**bar
			) for bar in data.pop('bars')
		]

		self.beats = [
			TimeInterval(
				start=timedelta(seconds=beat.pop('start')),
				duration=timedelta(seconds=beat.pop('duration')),
				**beat
			) for beat in data.pop('beats')
		]

		self.tatums = [
			TimeInterval(
				start=timedelta(seconds=tatum.pop('start')),
				duration=timedelta(seconds=tatum.pop('duration')),
				**tatum
			) for tatum in data.pop('tatums')
		]

		self.sections = [
			Section(
				start=timedelta(seconds=section.pop('start')),
				duration=timedelta(seconds=section.pop('duration')),
				**section
			) for section in data.pop('sections')
		]

		self.segments = [
			Segment(
				start=timedelta(seconds=segment.pop('start')),
				duration=timedelta(seconds=segment.pop('duration')),
				loudness_max_time=timedelta(seconds=segment.pop('loudness_max_time')),
				loudness_end=segment.pop('loudness_end', None),  # can be None, apparently
				**segment
			) for segment in data.pop('segments')
		]

		track_data = data.pop('track')

		self.track = Track(
			duration=timedelta(seconds=track_data.pop('duration')),
			offset_seconds=timedelta(seconds=track_data.pop('offset_seconds')),
			window_seconds=timedelta(seconds=track_data.pop('window_seconds')),
			end_of_fade_in=timedelta(seconds=track_data.pop('end_of_fade_in')),
			start_of_fade_out=timedelta(seconds=track_data.pop('start_of_fade_out')),
			**track_data
		)



from datetime import timedelta

from .object import SpotifyObject


class AudioFeatures(SpotifyObject):
	'''
	Represents an Audio Features object.

	id: str
		The Spotify ID of the track.
	uri: str
		Spotify URI of the album.
	analysis_url: str
		An HTTP URL to access the full audio analysis of this track.
	track_href: str
		A link to the Web API endpoint providing full details of the track.
	duration: timedelta
		The duration of the track.
	key: int
		The estimated overall key of the track.
	mode: int
		Mode indicates the modality (major or minor) of a track, the type of scale from which its melodic content is derived.
	time_signature: int
		An estimated overall time signature of a track.
	acousticness: float
		A confidence measure from 0.0 to 1.0 of whether the track is acoustic.
	danceability: float
		A measure of how suitable the track is for dancing.
	energy: float
		Energy is a measure from 0.0 to 1.0 and represents a perceptual measure of intensity and activity.
	instrumentalness: float
		Predicts whether a track contains no vocals.
	liveness: float
		Detects the presence of an audience in the recording.
	loudness: float
		The overall loudness of a track in decibels (dB).
	speechiness: float
		Speechiness detects the presence of spoken words in a track.
	valence: float
		A measure from 0.0 to 1.0 describing the musical positiveness conveyed by a track.
	tempo: float
		The overall estimated tempo of a track in beats per minute (BPM).
	'''

	_type = 'audio_features'

	def __init__(self, client, data):
		super().__init__(client, data)

		self.acousticness = data.pop('acousticness')
		self.analysis_url = data.pop('analysis_url')
		self.danceability = data.pop('danceability')
		self.duration = timedelta(milliseconds=data.pop('duration_ms'))
		self.energy = data.pop('energy')
		self.instrumentalness = data.pop('instrumentalness')
		self.key = data.pop('key')
		self.liveness = data.pop('liveness')
		self.loudness = data.pop('loudness')
		self.mode = data.pop('mode')
		self.speechiness = data.pop('speechiness')
		self.tempo = data.pop('tempo')
		self.time_signature = data.pop('time_signature')
		self.track_href = data.pop('track_href')
		self.valence = data.pop('valence')



import logging
from json import JSONDecodeError, dumps, loads
from os.path import isfile
from urllib.parse import parse_qs, urlparse

from .mixins import RefreshableFlowMixin
from .response import AuthenticationResponse, AuthorizationCodeFlowResponse
from ..http import Route
from ..scope import Scope

log = logging.getLogger(__name__)

AUTHORIZE_URL = 'https://accounts.spotify.com/authorize'
TOKEN_URL = 'https://accounts.spotify.com/api/token'


class Authenticator:
	'''Base authentication class. All authenticators should inherit from this, custom or not.'''

	_client = None
	_data: AuthenticationResponse = None

	def __init__(self, client_id, client_secret=None):
		self.client_id = client_id
		self.client_secret = client_secret

	def __call__(self, client):
		self._client = client
		return self

	async def close(self):
		pass

	@property
	def header(self):
		if self._data is None:
			raise ValueError('Not authorized yet, no Authorization header to craft')
		return self._data.header

	@property
	def client(self):
		if self._client is None:
			raise ValueError('Client not set for authenticator yet.')
		return self._client

	@property
	def market(self):
		raise RuntimeError('This authenticator does not ensure a market.')

	async def authorize(self):
		raise NotImplementedError


class ClientCredentialsFlow(Authenticator, RefreshableFlowMixin):
	'''
	Implements the Client Credentials flow.

	You can only access public resources using this authenticator.
	'''

	def __init__(self, client_id, client_secret, response_class=AuthenticationResponse):
		super().__init__(client_id, client_secret)
		self.response_class = response_class

	@property
	def market(self):
		# TODO: ??? does this make sense to do? I have no idea.
		return 'US'

	async def authorize(self):
		'''Authorize using this authenticator.'''

		await self.refresh(start_task=True)

	async def token(self):
		d = dict(
			grant_type='client_credentials',
			client_id=self.client_id,
			client_secret=self.client_secret
		)

		data = await self._token(d)

		return AuthenticationResponse(data)


class AuthorizationCodeFlow(Authenticator, RefreshableFlowMixin):
	'''
	Implements the Authorization Code flow.

	.. note::
	   This class is not for general use, please use :class:`EasyAuthorizationCodeFlow` or subclass this and implement
	   your own load(), store(response) and setup() methods.

	client_id: str
		Your application client id.
	client_secret: str
		Your application client secret.
	scope: :class:`Scope`
		The scope you're requesting.
	redirect_uri: str
		Where the user will be redirected to after accepting the client.
	response_class:
		The type that is expected to be returned from load() and setup(), and is passed to store(response) when a token refresh happens.
		Should be :class:`AuthorizationCodeFlowResponse` or inherit from it.
	'''

	_data: AuthorizationCodeFlowResponse

	def __init__(self, client_id, client_secret, scope, redirect_uri, response_class=AuthorizationCodeFlowResponse):
		super().__init__(client_id, client_secret)

		assert isinstance(scope, Scope)

		self.scope = scope
		self.redirect_uri = redirect_uri
		self.response_class = response_class

	@property
	def market(self):
		return 'from_token'

	async def token(self):
		if self._data is None:
			raise ValueError('Can\'t refresh token without previous refresh token')

		data = dict(
			grant_type='refresh_token',
			refresh_token=self._data.refresh_token,
			client_id=self.client_id,
			client_secret=self.client_secret
		)

		data = await self._token(data)

		ins = self.response_class(data)
		ins.refresh_token = self._data.refresh_token

		return ins

	async def authorize(self):
		'''Authorize the client. Reads from the file specificed by `store`.'''

		data = await self.load()

		# no data found, run first time setup
		# get response class, pass it to .store
		if data is None:
			data = await self.setup()

			if isinstance(data, AuthenticationResponse):
				await self.store(data)

		if not isinstance(data, AuthenticationResponse):
			raise TypeError('setup() has to return an AuthenticationResponse')

		self._data = data

		# refresh it now if it's expired
		if self._data.is_expired():
			await self.refresh(start_task=True)
		else:
			# manually start refresh task if we didn't refresh on startup
			self.refresh_in(self._data.seconds_until_expire())

	async def setup(self):
		raise NotImplementedError

	async def store(self, response):
		raise NotImplementedError

	async def load(self):
		raise NotImplementedError

	def create_authorize_route(self):
		'''Craft the :class:`Route` for the user to use for authorizing the client.'''

		params = dict(
			client_id=self.client_id,
			redirect_uri=self.redirect_uri,
			response_type='code'
		)

		if self.scope is not None:
			params['scope'] = self.scope.string()

		return Route('GET', AUTHORIZE_URL, **params)

	def get_code_from_redirect(self, url):
		'''Extract the authorization code from the redirect uri.'''

		parsed = urlparse(url.strip())
		query = parsed.query

		if not query:
			raise ValueError('Unable to parse that redirect url')

		qs = parse_qs(query)
		if 'code' not in qs:
			raise ValueError('Redirect url seems to be missing code fragment')

		return qs['code'][0]

	def create_token_data_from_code(self, code):
		return dict(
			client_id=self.client_id,
			client_secret=self.client_secret,
			grant_type='authorization_code',
			code=code,
			redirect_uri=self.redirect_uri
		)


class EasyAuthorizationCodeFlow(AuthorizationCodeFlow):
	def __init__(self, client_id, client_secret, scope=Scope.none(), storage='secret.json', response_class=AuthorizationCodeFlowResponse):
		super().__init__(client_id, client_secret, scope, 'http://localhost/', response_class)
		self.storage = storage

	async def setup(self):
		fmt = (
			'Hi! This is the initial EasyAuthorizationCode setup.\n\n'
			'Please open this URL:\n{0}\n\n'
			'and then input the URL you were redirected to after accepting here:\n'
		).format(str(self.create_authorize_route()))

		code_url = input(fmt)

		code = self.get_code_from_redirect(code_url)
		d = self.create_token_data_from_code(code)

		data = await self._token(d)
		return self.response_class(data)

	async def load(self):
		if isfile(self.storage):
			# if storage file exists, read and deserialize it
			with open(self.storage, 'r') as f:
				try:
					raw_data = loads(f.read())
				except JSONDecodeError:
					return None

				# return the response instance
				return self.response_class.from_data(raw_data)

	async def store(self, response):
		# simply store the response as a dumped json dict
		with open(self.storage, 'w') as f:
			f.write(dumps(response.to_dict(), indent=2))



from .flags import Flag, Flags


class Scope(Flags):
	'''
	Flags representing Spotify scopes.

	ugc_image_upload
		Write access to user-provided images.
	user_modify_playback_state
		Write access to a user’s playback state.
	user_read_playback_state
		Read access to a user’s player state.
	user_read_currently_playing
		Read access to a user’s currently playing content.
	user_top_read
		Read access to a user's top artists and tracks.
	user_read_playback_position
		Read access to a user’s playback position in a content.
	user_read_recently_played
		Read access to a user’s recently played tracks.
	user_library_modify
		Write/delete access to a user's "Your Music" library.
	user_library_read
		Read access to a user's "Your Music" library.
	user_follow_modify
		Write/delete access to the list of artists and other users that the user follows.
	user_follow_read
		Read access to the list of artists and other users that the user follows.
	playlist_read_private
		Read access to user's private playlists.
	playlist_modify_public
		Write access to a user's public playlists.
	playlist_modify_private
		Write access to a user's private playlists.
	playlist_read_collaborative
		Include collaborative playlists when requesting a user's playlists.
	user_read_private
		Read access to user’s subscription details (type of user account).
	user_read_email
		Read access to user’s email address.
	app_remote_control
		Remote control playback of Spotify. This scope is currently available to Spotify iOS and Android SDKs.
	streaming
		Control playback of a Spotify track. This scope is currently available to the Web Playback SDK. The user must have a Spotify Premium account.
	'''

	ugc_image_upload = Flag(0)

	user_modify_playback_state = Flag(1)
	user_read_playback_state = Flag(2)
	user_read_currently_playing = Flag(3)

	user_top_read = Flag(4)
	user_read_playback_position = Flag(5)
	user_read_recently_played = Flag(6)

	user_library_modify = Flag(7)
	user_library_read = Flag(8)

	user_follow_modify = Flag(9)
	user_follow_read = Flag(10)

	playlist_read_private = Flag(11)
	playlist_modify_public = Flag(12)
	playlist_modify_private = Flag(13)
	playlist_read_collaborative = Flag(14)

	user_read_private = Flag(15)
	user_read_email = Flag(16)

	app_remote_control = Flag(17)
	streaming = Flag(18)

	def string(self):
		'''Get a string representation of the enabled scopes. Used when authenticating.'''

		return ' '.join(name.replace('_', '-') for name, value in self if value)



class SpotifyObject:
	'''
	Represents a generic Spotify Object.
	
	Attributes
	----------
	
	id: str
		Spotify ID of the object.
	name: str
		Name of the object.
	uri: str
		Spotify URI of the object.
		
	'''

	_type = None

	def __init__(self, client, data):
		self._client = client
		self.id = data.pop('id', None)
		self.name = data.pop('name', None)
		self.href = data.pop('href', None)
		self.uri = data.pop('uri', None)

	@property
	def type(self):
		return self._type

	def __eq__(self, other):
		return self.id == other.id

	def __ne__(self, other):
		return not self.__eq__(other)

	def __repr__(self):
		repr = self.__class__.__name__
		if self.id is not None:
			repr += ' id=\'{}\''.format(self.id)
		if self.name is not None:
			repr += ' name=\'{}\''.format(self.name)
		return '<%s>' % repr



def get(items, **kwargs):
	'''
	Get an item from a list of items.

	:param items: List or iterator containing :class:`Object` s
	:param kwargs: kwargs that should match with the objects attributes.
	:return: First item that matched.
	'''
	for item in items:
		if _is_match(item, kwargs):
			return item
	return None


def find(items, **kwargs):
	'''
	Same as :func:`get` except it returns a list of all matching items.
	
	:param items: List or iterator containing :class:`Object`
	:param kwargs: kwargs that should match with the objects attributes.
	:return: List[:class:`Object`]
	'''
	return list(filter(lambda item: _is_match(item, kwargs), items))


def _is_match(item, kwargs):
	for k, v in kwargs.items():
		if getattr(item, k, None) != v:
			return False
	return True


def subslice(iter, step):
	group = []

	for idx, item in enumerate(iter):
		if idx % step == 0:
			if group:
				yield group
				group.clear()

		group.append(item)

	if group:
		yield group



"""
A Python library for interacting asynchronously with the Spotify API
"""

__title__ = 'asyncspotify'
__author__ = 'RUNIE'
__version__ = '0.11.8'
__license__ = 'MIT'

import logging

from . import utils
from .album import FullAlbum, SimpleAlbum
from .artist import FullArtist, SimpleArtist
from .audioanalysis import AudioAnalysis
from .audiofeatures import AudioFeatures
from .client import Client
from .device import Device
from .exceptions import *
from .http import Route
from .image import Image
from .oauth import AuthorizationCodeFlow, ClientCredentialsFlow, EasyAuthorizationCodeFlow
from .object import SpotifyObject
from .playing import CurrentlyPlaying, CurrentlyPlayingContext
from .playlist import FullPlaylist, SimplePlaylist
from .scope import Scope
from .track import FullTrack, PlaylistTrack, SimpleTrack
from .user import PrivateUser, PublicUser

log = logging.getLogger(__name__)
log.setLevel(logging.WARNING)



import asyncio
import logging
from asyncio import Lock, sleep
from json import JSONDecodeError, loads
from urllib.parse import urlencode

from aiohttp import ClientSession

from .exceptions import *

log = logging.getLogger(__name__)


def filter_nones(kwargs):
	return {k: v for k, v in kwargs.items() if v is not None}


class Route:
	BASE = 'https://api.spotify.com/v1'

	def __init__(self, method: str, url: str, **params):
		self.method = method.upper()
		self.url = url if url.startswith('http') else '{0}/{1}'.format(self.BASE, url)
		self.params = {k: v for k, v in params.items() if v is not None}

	def __repr__(self):
		return '<Route {0.method} url={0.url} params={0.params}>'.format(self)

	def __str__(self):
		ret = self.url
		if self.params:
			ret += '?' + urlencode(self.params)
		return ret


class HTTP:
	_attempts = 5

	def __init__(self, client, loop=None):
		self.client = client
		self.session = ClientSession(loop=loop or asyncio.get_event_loop())
		self.lock = Lock()

	async def close(self):
		await self.session.close()

	async def request(self, route, data=None, json=None, headers=None, authorize=True):
		if authorize:
			auth_header = self.client.auth.header
			if auth_header is None:
				raise AuthenticationError('Authorize before attempting an authorized request.')

			if headers:
				headers.update(auth_header)
			else:
				headers = auth_header

		kw = dict(method=route.method, url=route.url, headers=headers)

		if route.params:
			kw['params'] = route.params

		if data:
			kw['data'] = data

		if json:
			kw['json'] = json
			kw['headers']['Content-Type'] = 'application/json'

		async with self.lock:
			for attempt in range(self._attempts):

				async with self.session.request(**kw) as r:
					status_code = r.status
					headers = r.headers
					text = await r.text()

					log.debug('[%s] %s', status_code, repr(route))

					try:
						data = loads(text)
					except JSONDecodeError:
						data = None

					if 200 <= status_code < 300:
						return data

					try:
						error = data['error']['message']
					except (TypeError, KeyError):
						error = None

					if status_code == 429:
						retry_after = int(headers.get('Retry-After', 1)) + 1
						log.warning('Rate limited. Retrying in %s seconds.', retry_after)
						await sleep(retry_after)
						continue

					elif status_code == 400:
						raise BadRequest(r, error)

					elif status_code == 401:
						raise Unauthorized(r, error)

					elif status_code == 403:
						raise Forbidden(r, error)

					elif status_code == 404:
						raise NotFound(r, error)

					elif status_code == 405:
						raise NotAllowed(r, error)

					elif status_code >= 500:
						continue

					else:
						raise HTTPException(r, 'Unhandled HTTP status code: %s' % status_code)

		raise HTTPException(r, 'Request failed 5 times.')

	async def get_player(self, **kwargs):
		r = Route('GET', 'me/player', **kwargs)
		return await self.request(r)

	async def player_currently_playing(self, **kwargs):
		r = Route('GET', 'me/player/currently-playing', **kwargs)
		return await self.request(r)

	async def get_devices(self):
		r = Route('GET', 'me/player/devices')
		return await self.request(r)

	async def player_next(self, device_id):
		r = Route('POST', 'me/player/next', device=device_id)
		await self.request(r)

	async def player_prev(self, device_id):
		r = Route('POST', 'me/player/previous', device=device_id)
		await self.request(r)

	async def player_play(self, device_id, **kwargs):
		r = Route('PUT', 'me/player/play', device=device_id)
		await self.request(r, json=kwargs)

	async def player_pause(self, device_id):
		r = Route('PUT', 'me/player/pause', device=device_id)
		await self.request(r)

	async def player_seek(self, time, device_id):
		query = dict(position_ms=time)
		if device_id is not None:
			query['device'] = device_id
		r = Route('PUT', 'me/player/seek', **query)
		await self.request(r)

	async def player_repeat(self, state, device_id):
		query = dict(state=state)
		if device_id is not None:
			query['device'] = device_id
		r = Route('PUT', 'me/player/repeat', **query)
		await self.request(r)

	async def player_volume(self, volume, device_id):
		query = dict(volume_percent=volume)
		if device_id is not None:
			query['device'] = device_id
		r = Route('PUT', 'me/player/volume', **query)
		await self.request(r)

	async def player_shuffle(self, state, device_id):
		query = dict(state='true' if state else 'false')

		if device_id is not None:
			query['device'] = device_id

		r = Route('PUT', 'me/player/shuffle', **query)
		await self.request(r)

	async def search(self, q, type, **kwargs):
		r = Route('GET', 'search', type=type, q=q, **kwargs)
		return await self.request(r)

	async def get_me(self):
		r = Route('GET', 'me')
		return await self.request(r)

	async def get_me_top_tracks(self, **kwargs):
		r = Route('GET', 'me/top/tracks', **kwargs)
		return await self.request(r)

	async def get_me_top_artists(self, **kwargs):
		r = Route('GET', 'me/top/artists', **kwargs)
		return await self.request(r)

	async def get_playlist(self, playlist_id):
		r = Route('GET', 'playlists/{0}'.format(playlist_id))
		return await self.request(r)

	async def get_me_playlists(self):
		raise NotImplemented

	async def get_user_playlists(self, user_id):
		r = Route('GET', 'users/{0}/playlists'.format(user_id), limit=50)
		return await self.request(r)

	async def get_playlist_tracks(self, playlist_id):
		r = Route('GET', 'playlists/{0}/tracks'.format(playlist_id))
		return await self.request(r)

	async def playlist_add_tracks(self, playlist_id, **kwargs):
		data = dict(**filter_nones(kwargs))

		r = Route('POST', 'playlists/{0}/tracks'.format(playlist_id))
		return await self.request(r, json=data)

	async def create_playlist(self, user_id, name, **kwargs):
		data = dict(name=name, **filter_nones(kwargs))

		r = Route('POST', 'users/{0}/playlists'.format(user_id))
		return await self.request(r, json=data)

	async def edit_playlist(self, playlist_id, **kwargs):
		data = filter_nones(kwargs)

		r = Route('PUT', 'playlists/{0}'.format(playlist_id))
		await self.request(r, json=data)

	async def get_user(self, user_id):
		r = Route('GET', 'users/{0}'.format(user_id))
		return await self.request(r)

	async def get_track(self, track_id):
		r = Route('GET', 'tracks/{0}'.format(track_id))
		return await self.request(r)

	async def get_tracks(self, track_ids):
		r = Route('GET', 'tracks', ids=track_ids)
		return await self.request(r)

	async def get_audio_features(self, track_id):
		r = Route('GET', 'audio-features/{0}'.format(track_id))
		return await self.request(r)

	async def get_audio_analysis(self, track_id):
		r = Route('GET', 'audio-analysis/{0}'.format(track_id))
		return await self.request(r)

	async def get_artist(self, artist_id):
		r = Route('GET', 'artists/{0}'.format(artist_id))
		return await self.request(r)

	async def get_artist_albums(self, artist_id, **kwargs):
		req = Route('GET', 'artists/{0}/albums'.format(artist_id), **kwargs)
		return await self.request(req)

	async def get_artists(self, artist_ids):
		r = Route('GET', 'artists', ids=artist_ids)
		return await self.request(r)

	async def get_artist_top_tracks(self, artist_id, **kwargs):
		r = Route('GET', 'artists/{0}/top-tracks'.format(artist_id), **kwargs)
		return await self.request(r)

	async def get_artist_related_artists(self, artist_id):
		r = Route('GET', 'artists/{0}/related-artists'.format(artist_id))
		return await self.request(r)

	async def get_album(self, album_id, **kwargs):
		r = Route('GET', 'albums/{0}'.format(album_id), **kwargs)
		return await self.request(r)

	async def get_albums(self, album_ids, **kwargs):
		r = Route('GET', 'albums', ids=album_ids, **kwargs)
		return await self.request(r)

	async def get_album_tracks(self, album_id, **kwargs):
		r = Route('GET', 'albums/{0}/tracks'.format(album_id), limit=50, **kwargs)
		return await self.request(r)

	async def get_followed_artists(self, type, **kwargs):
		r = Route('GET', 'me/following', type=type, **kwargs)
		return await self.request(r)

	async def following(self, type, ids, **kwargs):
		r = Route('PUT', 'me/following', type=type, ids=ids, **kwargs)
		await self.request(r)



class SpotifyException(Exception):
	'''
	Base exception of all exceptions thrown by this library.
	'''
	pass


class AuthenticationError(SpotifyException):
	'''
	General exception when authenticating failed.
	'''


class HTTPException(SpotifyException):
	'''
	Base exception of all HTTP related exceptions.

	response: aiohttp.ClientResponse
		The response of the failed HTTP request.
	message: Optional[str]
		Message about what went wrong.

	'''

	def __init__(self, response, message=None):
		self.response = response
		self.message = message

		error = '{0.status} {0.reason}'.format(response)

		if message is not None:
			error += ': {}'.format(self.message)

		super().__init__(error)


class BadRequest(HTTPException):
	'''
	400 Bad Request
	
	Base class for all 4xx status code exceptions.
	'''
	pass


class Unauthorized(BadRequest):
	'''
	401 Unauthorized
	'''
	pass


class Forbidden(BadRequest):
	'''
	403 Forbidden
	'''
	pass


class NotFound(BadRequest):
	'''
	404 Not Found
	'''
	pass


class NotAllowed(BadRequest):
	'''
	405 Method Not Allowed
	'''
	pass



import logging

from .http import Route

log = logging.getLogger(__name__)


class Pager:
	def __init__(self, http, pager_object, limit=None):
		self.http = http
		self.pos = 0
		self.self_limit = limit
		self.set_next(pager_object)

	def set_next(self, obj):
		self.total = obj['total']
		self.next = obj['next']
		self.previous = obj['previous']
		self.items = obj['items']
		self.limit = obj['limit']
		self.offset = obj['offset']

	async def get_next(self):
		r = Route('GET', self.next)
		obj = await self.http.request(r)
		self.set_next(obj)

	def __aiter__(self):
		return self

	async def __anext__(self):
		# stop if we hit the pager total or the specified pager limit
		if self.pos >= self.total:
			raise StopAsyncIteration

		if self.self_limit is not None and self.pos >= self.self_limit:
			raise StopAsyncIteration

		# get the next page if we're exhausted this one
		if self.pos >= self.offset + self.limit:
			await self.get_next()

		item = self.items[self.pos - self.offset]
		self.pos += 1
		return item


class SearchPager(Pager):
	def __init__(self, http, obj, type, limit=None):
		self.type = type
		super().__init__(http, obj, limit)

	def set_next(self, obj):
		obj = obj[self.type]
		super().set_next(obj)


class CursorBasedPaging(SearchPager):
	def set_next(self, obj):
		obj[self.type]['offset'] = None
		self.cursors = obj[self.type]['cursors']
		super().set_next(obj)



from datetime import datetime, timedelta

from .device import Device
from .object import SpotifyObject
from .track import FullTrack


class CurrentlyPlaying(SpotifyObject):
	'''
	Represents a Currently Playing object.

	timestamp: datetime
		When the object information was created by the Spotify API.
	progress: timedelta
		How far into the current track the player is.
	is_playing: bool
		Whether the track is playing or not.
	track: :class:`Track`
		What track is currently playing, can be ``None``
	currently_playing_type: str
		What is currently playing, can be ``track``, ``episode``, ``ad`` or ``unknown``.
	'''

	def __init__(self, client, data):
		super().__init__(client, data)

		self.timestamp = datetime.utcfromtimestamp(data.pop('timestamp') / 1000)
		self.progress = timedelta(milliseconds=data.pop('progress_ms'))
		self.is_playing = data.pop('is_playing')

		item = data.pop('item', None)
		self.track = None if item is None else FullTrack(client, item)

		self.currently_playing_type = data.pop('currently_playing_type')


class CurrentlyPlayingContext(CurrentlyPlaying):
	'''
	Represents a Player object, extends :class:`CurrentlyPlaying`

	This type has some additional attributes not existent in :class:`CurrentlyPlaying`.

	device: :class:`Device`
		What device is owns this context.
	repeat_state: str
		The repeat state of the player. Can be ``off``, ``track`` or ``context``.
	shuffe_state: bool
		The shuffle state of the player. Can be ``True`` or ``False``.
	'''

	def __init__(self, client, data):
		super().__init__(client, data)

		self.device = Device(client, data.pop('device'))
		self.repeat_state = data.pop('repeat_state')
		self.shuffle_state = data.pop('shuffle_state')

	async def next(self):
		'''Skips to the next track.'''
		await self._client.player_next(device=self.device.id)

	async def prev(self):
		'''Goes to the previous track.'''
		await self._client.player_prev(device=self.device.id)

	async def play(self, **kwargs):
		'''
		Starts playback.

		:param kwargs: Body parameters of the request.

		.. code-block:: py

			player_play(
				context_uri='spotify:album:1Je1IMUlBXcx1Fz0WE7oPT',
				offset=dict(uri='spotify:track:1301WleyT98MSxVHPZCA6M'),
				position_ms=1000
			)
		'''

		await self._client.player_play(device=self.device.id, **kwargs)

	async def pause(self):
		'''Pauses playback.'''
		await self._client.player_pause(device=self.device.id)

	async def seek(self, time):
		'''
		Seeks to a specified time in the current track.

		:param time: timedelta object or milliseconds (integer)
		'''

		await self._client.player_seek(time=time, device=self.device.id)

	async def repeat(self, state):
		'''
		Set player repeat mode.

		:param str state: Can be 'track', 'context' or 'off'.
		'''

		await self._client.player_repeat(state=state, device=self.device.id)

	async def volume(self, volume):
		'''
		Set player volume.

		:param int volume: Value from 0 to 100.
		'''
		await self._client.player_volume(volume=volume, device=self.device.id)

	async def shuffle(self, state):
		'''
		Set player shuffle mode.

		:param bool state: Shuffle mode state.
		'''

		await self._client.player_shuffle(state=state, device=self.device.id)



class Flag:
	def __init__(self, pos: int):
		self.mask = 1 << pos

	def __set__(self, obj, val):
		return obj._set_flag(self.mask, val)

	def __get__(self, instance, owner):
		return instance._get_flag(self.mask)


class Flags:
	def __init__(self, value=0, **kwargs):
		self._value = value

		if kwargs:
			for name, val in self:
				new_val = kwargs.pop(name, None)
				if new_val is not None:
					setattr(self, name, new_val)

		if kwargs:
			raise ValueError('kwarg(s) does not have an associated flag: {0}'.format(', '.join(kwargs)))

	def __repr__(self):
		return '<%s value=%s>' % (self.__class__.__name__, self._value)

	def __iter__(self):
		for attr, flag in self.__class__.__dict__.items():
			if isinstance(flag, Flag):
				yield attr, self._get_flag(flag.mask)

	@classmethod
	def all(cls):
		obj = cls()
		for name, value in obj:
			setattr(obj, name, True)
		return obj

	@classmethod
	def none(cls):
		return cls(0b0)

	def _set_flag(self, mask: int, toggle):
		if toggle:
			self._value |= mask
		else:
			self._value &= ~mask

	def _get_flag(self, mask):
		return bool(self._value & mask)



from datetime import datetime

from .mixins import ArtistMixin, ExternalIDMixin, ExternalURLMixin, ImageMixin, TrackMixin
from .object import SpotifyObject
from .track import SimpleTrack


class _BaseAlbum(SpotifyObject, TrackMixin, ImageMixin, ExternalURLMixin, ArtistMixin):
	_type = 'album'
	_track_class = SimpleTrack
	__date_fmt = dict(year='%Y', month='%Y-%m', day='%Y-%m-%d')

	def __init__(self, client, data):
		super().__init__(client, data)

		TrackMixin.__init__(self, data)
		ImageMixin.__init__(self, data)
		ExternalURLMixin.__init__(self, data)
		ArtistMixin.__init__(self, data)

		self.album_group = data.pop('album_group', None)  # can be None, though this is not specified in the API docs
		self.album_type = data.pop('album_type')
		self.available_markets = data.pop('available_markets', None)

		self.release_date_precision = data.pop('release_date_precision')

		if self.release_date_precision is None:
			self.release_date = None
		else:
			try:
				self.release_date = datetime.strptime(
					data.pop('release_date'),
					self.__date_fmt[self.release_date_precision]
				)
			except ValueError:
				self.release_date = None


class SimpleAlbum(_BaseAlbum):
	'''
	Represents an Album object.

	.. note::
	   To iterate all tracks, you have to use the ``async for`` construct or fill the object with ``.fill()`` before iterating ``.tracks``.

	id: str
		Spotify ID of the album.
	name: str
		Name of the album.
	tracks: List[:class:`Track`]
		List of tracks on the album.
	artists: List[:class:`Artist`]
		List of artists that appear on the album.
	images: List[:class:`Image`]
		List of associated images, such as album cover in different sizes.
	uri: str
		Spotify URI of the album.
	link: str
		Spotify URL of the album.
	type: str
		Plaintext string of object type: ``album``.
	album_type:
		Type of album, e.g. ``album``, ``single`` or ``compilation``.
	available_markets: List[str] or None
		Markets where the album is available: ISO-3166-1_.
	external_urls: dict
		Dictionary that maps type to url.
	release_date: `datetime <https://docs.python.org/3/library/datetime.html#module-datetime>`_
		Date (and maybe time) of album release.
	release_date_precision: str
		Precision of ``release_date``. Can be ``year``, ``month``, or ``day``.
	album_group: str or None
		Type of album, e.g. ``album``, ``single``, ``compilation`` or ``appears_on``.
	'''


class FullAlbum(_BaseAlbum, ExternalIDMixin):
	'''
	Represents a complete Album object.
	
	This type has some additional attributes not existent in :class:`SimpleAlbum`.
	
	genres: List[str]
		List of genres associated with the album.
	label: str
		The label for the album.
	popularity: int
		An indicator of the popularity of the album, 0 being least popular and 100 being the most.
	copyrights: dict
		List of copyright objects.
	external_ids: dict
		Dictionary of external IDs.
	'''

	def __init__(self, client, data):
		super().__init__(client, data)

		ExternalIDMixin.__init__(self, data)

		self.genres = data.pop('genres')
		self.label = data.pop('label')
		self.popularity = data.pop('popularity')
		self.copyrights = data.pop('copyrights')



from .object import SpotifyObject


class Device(SpotifyObject):
	'''
	Represents a Device object.

	is_active: bool
		Whether this device is currently the active device.
	is_private_session: bool
		If the device session is private.
	is_restricted: bool
		Whether controlling this device is restricted. If this is ``true``, no API commands will work on it.
	name: str
		Name of this device.
	type: str
		Equal to ``device``.
	volume_percent: int
		Volume of this device. Integer between 0 to 100.
	'''

	_type = 'device'

	def __init__(self, client, data):
		super().__init__(client, data)

		self.is_active = data.pop('is_active')
		self.is_private_session = data.pop('is_private_session')
		self.is_restricted = data.pop('is_restricted')
		self.volume_percent = data.pop('volume_percent')



class Image:
	'''
	Represents an image.

	url: str
		URL of the image.
	width: int
		Width of the image
	height: int
		Height of the image.
	'''

	def __init__(self, data):
		self.url = data.pop('url')
		self.width = data.pop('width')
		self.height = data.pop('height')

	def __str__(self):
		return self.url

	def __repr__(self):
		return '<Image url=\'{}\''.format(self.url)



from .mixins import ExternalURLMixin, ImageMixin, TrackMixin
from .object import SpotifyObject
from .user import PublicUser
from .track import PlaylistTrack


class _BasePlaylist(SpotifyObject, TrackMixin, ExternalURLMixin, ImageMixin):
	_type = 'playlist'
	_track_class = PlaylistTrack

	def __init__(self, client, data):
		super().__init__(client, data)

		TrackMixin.__init__(self, data)
		ExternalURLMixin.__init__(self, data)
		ImageMixin.__init__(self, data)

		self.snapshot_id = data.pop('snapshot_id')
		self.collaborative = data.pop('collaborative')
		self.public = data.pop('public')

		self.owner = PublicUser(client, data.pop('owner'))

	async def edit(self, name=None, public=None, collaborative=None, description=None):
		'''
		Edit the playlist.
		
		:param str name: New name of the playlist.
		:param str description: New description of the playlist.
		:param bool public: New public state of the playlist.
		:param bool collaborative: New collaborative state of the playlist.
		'''

		await self._client.edit_playlist(
			playlist=self.id,
			name=name,
			public=public,
			collaborative=collaborative,
			description=description,
		)

	async def add_track(self, track, position=None):
		'''
		Add a track to the playlist.

		:param track: Spotify ID or :class:`Track` instance.
		:param int position: Position in the playlist to insert tracks.
		'''

		await self._client.playlist_add_tracks(self.id, track, position=position)

	async def add_tracks(self, *tracks, position=None):
		'''
		Add several tracks to the playlist.

		:param tracks: Several Spotify IDs or :class:`Track` instances (or a mix).
		:param int position: Position in the playlist to insert tracks.
		'''

		await self._client.playlist_add_tracks(self.id, *tracks, position=position)


class SimplePlaylist(_BasePlaylist):
	'''
	Represents a playlist object.

	.. note::
	   To iterate all tracks, you have to use the ``async for`` construct or fill the object with ``.fill()`` before iterating ``.tracks``.

	id: str
		Spotify ID of the playlist.
	name: str
		Name of the playlist.
	tracks: List[:class:`SimpleTrack`]
		All tracks in the playlist.
	uri: str
		Spotify URI of the playlist.
	link: str
		Spotify URL of the playlist.
	snapshot_id: str
		Spotify ID of the current playlist snapshot. Read about snapshots `here. <https://developer.spotify.com/documentation/general/guides/working-with-playlists/>`_
	collaborative: bool
		Whether the playlist is collaborative.
	public: bool
		Whether the playlist is public.
	owner: :class:`PublicUser`
		Owner of the playlist.
	external_urls: dict
		Dictionary that maps type to url.
	images: List[:class:`Image`]
		List of associated images.
	'''


class FullPlaylist(_BasePlaylist):
	'''
	Represents a complete playlist object.
	
	This type has some additional attributes not existent in :class:`SimplePlaylist`.
	
	description: str
		Description of the playlist, as set by the owner.
	primary_color: str
		Primary color of the playlist, for aesthetic purposes.
	follower_count: int
		Follower count of the playlist.
	'''

	def __init__(self, client, data):
		super().__init__(client, data)

		self.description = data.pop('description')
		self.primary_color = data.pop('primary_color')
		self.follower_count = data['followers']['total']



import setuptools

with open('README.rst', 'r') as fh:
    long_description = fh.read()

with open('requirements.txt', 'r') as f:
    install_requires = f.read().splitlines()

setuptools.setup(
    name="asyncspotify",
    version="0.11.8",
    author="RUNIE",
    author_email="runar-borge@hotmail.com",
    description="Spotify Web API implementation that is fully asynchronous and object-oriented.",
    long_description=long_description,
    long_description_content_type="text/x-rst",
    keywords='spotify async aio asyncio api webapi',
    install_requires=install_requires,
    url="https://github.com/Run1e/asyncspotify",
    project_urls={
        'Documentation': 'https://asyncspotify.rtfd.io/'
    },
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.5.2"  # first version to support async class generator iterators
)


import asyncio
import logging

from ..http import Route
from ..exceptions import HTTPException

log = logging.getLogger(__name__)


class RefreshableFlowMixin:
	_task: asyncio.Task = None

	async def refresh(self, start_task=True):
		'''Refresh this authenticator.'''

		meth = getattr(self, 'token', None)

		if not callable(meth):
			raise ValueError('This authorizer does not support token refreshing')

		data = None

		try:
			data = await meth()
			self._data = data
			if hasattr(self, 'store'):
				await self.store(data)
		finally:
			if start_task:
				if data is None:
					raise RuntimeError('Can\'t restart token refresh task when previous refresh failed')
				self.refresh_in(data.seconds_until_expire())

	async def _token(self, data):
		data = await self.client.http.request(
			Route('POST', 'https://accounts.spotify.com/api/token'),
			data=data,
			authorize=False
		)

		# data['expires_in'] = 5

		return data

	def refresh_in(self, seconds, cancel_task=True):
		# cancel old task if it's running
		if cancel_task and self._task is not None:
			self._task.cancel()

		# and create new refresh task
		self._task = asyncio.create_task(self._refresh_in_meta(seconds))

	async def _refresh_in_meta(self, seconds):
		if seconds > 0:
			log.info('Refreshing access token in %s seconds', seconds)
			await asyncio.sleep(seconds)
			log.debug('Refreshing access token now', seconds)

		# TODO: this back-off thing is pretty eh
		# it's not really even backing off
		# maybe also do a check to see whether the data is *actually* updated?
		try:
			await self.refresh(start_task=False)
		except HTTPException as e:
			log.critical(e, exc_info=True)
			next_attempt = 5.0
		else:
			next_attempt = self._data.seconds_until_expire()

		self.refresh_in(next_attempt, cancel_task=False)

	async def close(self):
		if self._task is not None:
			try:
				self._task.cancel()
				await self._task
			except asyncio.CancelledError:
				pass
